from typing import Sequence

import jax
import numpy as np
import pysindy as ps
import pytest
import sympy as sp
from sympy.parsing.sympy_parser import T

XFORMS = T[:6]
jax.config.update("jax_platform_name", "cpu")  # diffrax issue 722
# Alsorequires successive E402


from sindy_exp._data import gen_data  # noqa: E402
from sindy_exp._plotting import _coeff_dicts_to_matrix  # noqa: E402
from sindy_exp._typing import NestedDict  # noqa: E402
from sindy_exp._utils import unionize_coeff_dicts  # noqa: E402
from sindy_exp._utils import (  # noqa: E402
    _sindy_equations_to_sympy,
    _sympy_expr_to_feat_coeff,
)


@pytest.fixture
def jax_cpu_only():
    with jax.default_device(jax.devices("cpu")[0]):
        yield


def test_flatten_nested_dict():
    deep = NestedDict(a=NestedDict(b=1))
    result = deep.flatten()
    assert deep != result
    expected = {"a.b": 1}
    assert result == expected


def test_flatten_nested_bad_dict():
    nested = {1: NestedDict(b=1)}
    # Testing the very thing that causes a typing error, thus ignoring
    with pytest.raises(TypeError, match="keywords must be strings"):
        NestedDict(**nested)  # type: ignore
    with pytest.raises(TypeError, match="Only string keys allowed"):
        deep = NestedDict(a={1: 1})
        deep.flatten()


# Common input feature naming conventions
@pytest.mark.parametrize(
    "sys_coords",
    [("x", "y"), ("x_1", "x_2"), ("x1", "x2"), ("ab", "uv")],
    ids=["simple", "underscore", "numbered", "arbitrary"],
)
def test_sindy_equations_to_sympy(sys_coords):
    f1, f2 = sys_coords
    model = ps.SINDy().fit(np.random.normal(size=(10, 2)), 1, feature_names=[f1, f2])
    opt = model.optimizer
    opt.coef_ = np.random.normal(size=opt.coef_.shape)
    _, sympy_eqs = _sindy_equations_to_sympy(model)
    coeff_dict = _sympy_expr_to_feat_coeff(sympy_eqs)
    expected_keys = {
        sp.Integer(1.0),
        sp.Symbol(f1),
        sp.Symbol(f2),
        sp.Mul(sp.Symbol(f1), sp.Symbol(f2)),
        sp.Pow(sp.Symbol(f1), 2),
        sp.Pow(sp.Symbol(f2), 2),
    }
    assert all(d.keys() == expected_keys for d in coeff_dict)


# Consider additional feature libraries (e.g. cos, sin) later
@pytest.mark.parametrize(
    "str_exprs, expected_coeff_dict",
    [
        pytest.param(["3"], {sp.Integer(1.0): 3.0}, id="single_const_term"),
        pytest.param(["3x"], {sp.Symbol("x"): 3.0}, id="single_term"),
        pytest.param(["3.1 x"], {sp.Symbol("x"): 3.1}, id="single_term_alt"),
        pytest.param(["z"], {sp.Symbol("z"): 1.0}, id="single_term_no_coeff"),
        pytest.param(
            ["4xy"],
            {sp.Mul(sp.Symbol("x"), sp.Symbol("y")): 4.0},
            id="single_term_mult",
        ),
        pytest.param(
            ["xy"],
            {sp.Mul(sp.Symbol("x"), sp.Symbol("y")): 1.0},
            id="single_term_mult_no_coeff",
        ),
        pytest.param(
            ["2*x*y"],
            {sp.Mul(sp.Symbol("x"), sp.Symbol("y")): 2.0},
            id="single_term_mult_alt",
        ),
        pytest.param(["3x**2"], {sp.Pow(sp.Symbol("x"), 2): 3.0}, id="single_term_pow"),
        pytest.param(
            ["x**2"], {sp.Pow(sp.Symbol("x"), 2): 1.0}, id="single_term_pow_no_coeff"
        ),
        pytest.param(
            ["3x + x**2"],
            {sp.Symbol("x"): 3.0, sp.Pow(sp.Symbol("x"), 2): 1.0},
            id="multiple_terms",
        ),
        pytest.param(
            ["x**2 + 3x"],
            {sp.Pow(sp.Symbol("x"), 2): 1.0, sp.Symbol("x"): 3.0},
            id="multiple_terms_reverse_order",
        ),
    ],
)
def test_sympy_expr_to_feat_coeff(str_exprs, expected_coeff_dict):
    sp_exprs = [
        sp.parse_expr(ex, transformations=XFORMS, evaluate=False) for ex in str_exprs
    ]
    result = _sympy_expr_to_feat_coeff(sp_exprs)
    assert result == [expected_coeff_dict]
    assert list(result[0]) == list(expected_coeff_dict)


def test_unionize_coeff_dicts_aligns_features():
    class DummyFeatures:
        def get_feature_names(
            self, input_features: Sequence = ["x", "y", "z"]
        ) -> list[str]:
            x, y, z = tuple(input_features)
            return ["1", x, y, z]

    class DummyModel:
        feature_names: list[str] = ["x", "y", "z"]
        feature_library: ps.feature_library.base.BaseFeatureLibrary = DummyFeatures()

        def equations(self, precision: int = 10):
            # Three coordinates:
            return ["x", "x + y", "x + y"]

    x, y, z = sp.symbols("x y z")
    true_equations = [
        {x: 2.0},
        {x: 2.0, z: 3.0},
        {x: 2.0, z: 3.0},
    ]

    true_aligned, est_aligned = unionize_coeff_dicts(DummyModel(), true_equations)

    # All coordinates should share the same feature keys, in order
    all_keys = list(true_aligned[0].keys())
    assert all(list(d.keys()) == all_keys for d in true_aligned)
    assert all(list(d.keys()) == all_keys for d in est_aligned)

    # The feature union should include x, y, and z
    assert all_keys == [1, x, y, z]


def test_coeff_dicts_to_matrix_basic():
    x, y = sp.symbols("x y")
    coeffs = [
        {x: 1.0, y: 2.0},
        {x: 3.0, y: 4.0},
    ]

    mat, feature_names = _coeff_dicts_to_matrix(coeffs)

    # Shape matches number of coordinates and features
    assert mat.shape == (2, 2)

    # Features are ordered by their string representation
    assert feature_names == ["x", "y"]

    # Rows correspond to the input dictionaries
    assert mat[0, :].tolist() == [1.0, 2.0]
    assert mat[1, :].tolist() == [3.0, 4.0]


@pytest.mark.parametrize(
    "rhs_name", ["lorenz", "rossler", "vanderpol", "sho", "cubicho", "kinematics"]
)
@pytest.mark.parametrize("array_namespace", ["numpy", "jax"])
def test_gen_data(rhs_name, array_namespace, jax_cpu_only):
    result = gen_data(
        rhs_name, t_end=0.1, noise_abs=0.01, seed=42, array_namespace=array_namespace
    )["data"]
    trajectories = result["trajectories"]
    assert len(trajectories) == 1
    traj = trajectories[0]
    assert traj.x_train.shape == traj.x_train_true_dot.shape



from ._data import ODE_CLASSES, gen_data
from ._odes import fit_eval, plot_ode_panel
from ._plotting import (
    COLOR,
    compare_coefficient_plots_from_dicts,
    plot_coefficients,
    plot_test_trajectory,
    plot_training_data,
)
from ._typing import DynamicsTrialData, ProbData
from ._utils import coeff_metrics, integration_metrics, pred_metrics

__all__ = [
    "gen_data",
    "fit_eval",
    "ProbData",
    "DynamicsTrialData",
    "coeff_metrics",
    "pred_metrics",
    "integration_metrics",
    "ODE_CLASSES",
    "plot_ode_panel",
    "plot_coefficients",
    "compare_coefficient_plots_from_dicts",
    "plot_test_trajectory",
    "plot_training_data",
    "COLOR",
]



from dataclasses import dataclass
from typing import Annotated, Optional, Sequence
from warnings import warn

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy
import seaborn as sns
import sympy as sp
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.typing import ColorType

PAL = sns.color_palette("Set1")
PLOT_KWS = {"alpha": 0.7, "linewidth": 3}


@dataclass
class _ColorConstants:
    color_sequence: list[ColorType]

    def set_sequence(self, color_sequence: list[ColorType]):
        self.color_sequence = color_sequence

    @property
    def TRUE(self):
        return self.color_sequence[0]

    @property
    def MEAS(self):
        return self.color_sequence[1]

    @property
    def EST(self):
        return self.color_sequence[2]

    @property
    def SIM(self):
        return self.color_sequence[3]

    @property
    def TRAIN(self):
        return self.color_sequence[4]

    @property
    def TEST(self):
        return self.color_sequence[5]


COLOR = _ColorConstants(mpl.color_sequences["tab10"])


def plot_coefficients(
    coefficients: Annotated[np.ndarray, "(n_coord, n_features)"],
    input_features: Sequence[str],
    feature_names: Sequence[str],
    ax: Axes,
    **heatmap_kws,
) -> None:
    """Plot a set of dynamical system coefficients in a heatmap.

    Args:
        coefficients: A 2D array holding the coefficients of different
            library functions.  System dimension is rows, function index
            is columns
        input_features: system coordinate names, e.g. "x","y","z" or "u","v"
        feature_names: the names of the functions in the library.
        ax: the matplotlib axis to plot on
        **heatmap_kws: additional kwargs to seaborn's styling
    """

    def detex(input: str) -> str:
        if input[0] == "$":
            input = input[1:]
        if input[-1] == "$":
            input = input[:-1]
        return input

    if input_features is None:
        input_features = [r"$\dot x_" + f"{k}$" for k in range(coefficients.shape[0])]
    else:
        input_features = [r"$\dot " + f"{detex(fi)}$" for fi in input_features]

    if feature_names is None:
        feature_names = [f"f{k}" for k in range(coefficients.shape[1])]

    with sns.axes_style(style="white", rc={"axes.facecolor": (0, 0, 0, 0)}):
        heatmap_args = {
            "xticklabels": input_features,
            "yticklabels": feature_names,
            "center": 0.0,
            "cmap": sns.color_palette("vlag", n_colors=20, as_cmap=True),
            "ax": ax,
            "linewidths": 0.1,
            "linecolor": "whitesmoke",
        }
        heatmap_args.update(**heatmap_kws)
        coefficients = np.where(
            coefficients == 0, np.nan * np.empty_like(coefficients), coefficients
        )
        sns.heatmap(coefficients.T, **heatmap_args)

        ax.tick_params(axis="y", rotation=0)

    return ax


def _coeff_dicts_to_matrix(
    coeffs: Sequence[dict[sp.Expr, float]],
) -> tuple[np.ndarray, list[str]]:
    """Convert a list of coefficient dictionaries to a dense matrix.

    The input is a list of dictionaries mapping feature identifiers (
    SymPy expressions) to coefficients. All dictionaries are assumed to
    correspond to different coordinates of the same system. This helper builds
    a consistent feature ordering across coordinates and returns a numeric
    matrix along with the stringified feature names.
    """

    if not coeffs:
        raise ValueError("No coefficient dictionaries provided.")

    features: list[sp.Expr] = sorted({key for d in coeffs for key in d.keys()}, key=str)

    mat = np.zeros((len(coeffs), len(features)), dtype=float)
    for row, d in enumerate(coeffs):
        for col, feat in enumerate(features):
            mat[row, col] = d[feat]

    feature_names = [str(f).replace("**", "^") for f in features]
    return mat, feature_names


def _compare_coefficient_plots_impl(
    coefficients_est: Annotated[np.ndarray, "(n_coord, n_feat)"],
    coefficients_true: Annotated[np.ndarray, "(n_coord, n_feat)"],
    input_features: Sequence[str],
    feature_names: Sequence[str],
    scaling: bool = True,
    axs: Optional[Sequence[Axes]] = None,
) -> None:
    """Internal implementation for coefficient comparison heatmaps."""
    n_cols = len(coefficients_est)

    # helps boost the color of small coefficients.  Maybe log is better?
    all_vals = np.hstack((coefficients_est.flatten(), coefficients_true.flatten()))
    nzs = all_vals[all_vals.nonzero()]
    max_val = np.max(np.abs(nzs), initial=0.0)
    min_val = np.min(np.abs(nzs), initial=np.inf)
    if scaling and np.isfinite(min_val) and max_val / min_val > 10:
        pwr_ratio = 1.0 / np.log10(max_val / min_val)
    else:
        pwr_ratio = 1

    def signed_root(x):
        return np.sign(x) * np.power(np.abs(x), pwr_ratio)

    with sns.axes_style(style="white", rc={"axes.facecolor": (0, 0, 0, 0)}):
        if axs is None:
            fig, axs = plt.subplots(
                1, 2, figsize=(1.9 * n_cols, 8), sharey=True, sharex=True
            )
            fig.tight_layout()

        vmax = signed_root(max_val)

        plot_coefficients(
            signed_root(coefficients_true),
            input_features=input_features,
            feature_names=feature_names,
            ax=axs[0],
            cbar=False,
            vmax=vmax,
            vmin=-vmax,
        )

        plot_coefficients(
            signed_root(coefficients_est),
            input_features=input_features,
            feature_names=feature_names,
            ax=axs[1],
            cbar=False,
            vmax=vmax,
            vmin=-vmax,
        )

        axs[0].set_title("True Coefficients", rotation=45)
        axs[1].set_title("Est. Coefficients", rotation=45)


def compare_coefficient_plots(
    coefficients_est: Annotated[np.ndarray, "(n_coord, n_feat)"],
    coefficients_true: Annotated[np.ndarray, "(n_coord, n_feat)"],
    input_features: Sequence[str],
    feature_names: Sequence[str],
    scaling: bool = True,
    axs: Optional[Sequence[Axes]] = None,
) -> None:
    """Create plots of true and estimated coefficients.

    Deprecated:
        Use :func:`compare_coefficient_plots_from_dicts` with coefficient
        dictionaries instead. This function will be removed in a future
        release.
    """

    warn(
        "compare_coefficient_plots is deprecated; use "
        "compare_coefficient_plots_from_dicts with coefficient dictionaries instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    _compare_coefficient_plots_impl(
        coefficients_est,
        coefficients_true,
        input_features=input_features,
        feature_names=feature_names,
        scaling=scaling,
        axs=axs,
    )


def compare_coefficient_plots_from_dicts(
    coefficients_est: Sequence[dict[sp.Expr, float]],
    coefficients_true: Sequence[dict[sp.Expr, float]],
    input_features: Sequence[str],
    feature_names: Sequence[str] | None = None,
    scaling: bool = True,
    axs: Optional[Sequence[Axes]] = None,
):
    """Wrapper to compare coefficients given as dictionaries.

    Converts aligned coefficient dictionaries into dense matrices and then
    delegates to :func:`compare_coefficient_plots` for plotting.

    This assumes that the coefficient dictionaries are aligned, i.e., that they
    contain the same keys across all coordinates, as produced by
    ``unionize_coeff_dicts()``.
    """

    true_mat, inferred_feature_names = _coeff_dicts_to_matrix(coefficients_true)
    est_mat, est_feature_names = _coeff_dicts_to_matrix(coefficients_est)

    if true_mat.shape != est_mat.shape:
        raise ValueError("True and estimated coefficient shapes do not match")

    if inferred_feature_names != est_feature_names:
        raise ValueError(
            "Feature names inferred from true and estimated coefficients do not match"
        )

    _compare_coefficient_plots_impl(
        est_mat,
        true_mat,
        input_features=input_features,
        feature_names=inferred_feature_names,
        scaling=scaling,
        axs=axs,
    )


def _plot_training_trajectory(
    ax: Axes,
    x_train: np.ndarray,
    x_true: np.ndarray,
    x_smooth: np.ndarray | None,
    labels: bool = True,
) -> None:
    """Plot a single training trajectory

    If x_smooth is provided, it is only plotted if sufficiently different
    from x_train.
    """
    if x_train.shape[1] == 2:
        ax.plot(
            x_true[:, 0], x_true[:, 1], ".", label="True", color=COLOR.TRUE, **PLOT_KWS
        )
        ax.plot(
            x_train[:, 0],
            x_train[:, 1],
            ".",
            label="Measured",
            color=COLOR.MEAS,
            **PLOT_KWS,
        )
        if (
            x_smooth is not None
            and np.linalg.norm(x_smooth - x_train) / np.linalg.norm(x_train) > 1e-12
        ):
            ax.plot(
                x_smooth[:, 0],
                x_smooth[:, 1],
                ".",
                label="Smoothed",
                color=COLOR.EST,
                **PLOT_KWS,
            )
        if labels:
            ax.set(xlabel="$x_0$", ylabel="$x_1$")
        else:
            ax.set(xticks=[], yticks=[])
    elif x_train.shape[1] == 3:
        ax.plot(
            x_true[:, 0],
            x_true[:, 1],
            x_true[:, 2],
            color=COLOR.TRUE,
            label="True values",
            **PLOT_KWS,
        )

        ax.plot(
            x_train[:, 0],
            x_train[:, 1],
            x_train[:, 2],
            ".",
            color=COLOR.MEAS,
            label="Measured values",
            alpha=0.3,
        )
        if (
            x_smooth is not None
            and np.linalg.norm(x_smooth - x_train) / x_smooth.size > 1e-12
        ):
            ax.plot(
                x_smooth[:, 0],
                x_smooth[:, 1],
                x_smooth[:, 2],
                ".",
                color=COLOR.EST,
                label="Smoothed values",
                alpha=0.3,
            )
        if labels:
            ax.set(xlabel="$x$", ylabel="$y$", zlabel="$z$")
        else:
            ax.set(xticks=[], yticks=[], zticks=[])
    else:
        raise ValueError("Can only plot 2d or 3d data.")


def plot_training_data(
    t_train: np.ndarray,
    x_train: np.ndarray,
    x_true: np.ndarray,
    x_smooth: np.ndarray | None = None,
    coord_names: Optional[Sequence[str]] = None,
) -> tuple[Figure, Figure]:
    """Plot training data (and smoothed training data, if different)."""
    if coord_names is None:
        coord_names = [f"$x_{i}$" for i in range(x_true.shape[1])]

    fig_composite = plt.figure(figsize=(12, 6))
    if x_train.shape[-1] == 2:
        ax_traj = fig_composite.add_subplot(1, 2, 1)
    elif x_train.shape[-1] == 3:
        ax_traj = fig_composite.add_subplot(1, 2, 1, projection="3d")
    else:
        raise ValueError("Too many or too few coordinates to plot")
    _plot_training_trajectory(ax_traj, x_train, x_true, x_smooth)
    ax_traj.legend()
    ax_traj.set(title="Trajectory Plot")
    ax_psd = fig_composite.add_subplot(1, 2, 2)
    _plot_data_psd(ax_psd, x_train, coord_names, traj_type="train")
    _plot_data_psd(ax_psd, x_true, coord_names, traj_type="true")
    ax_psd.set(title="Absolute Spectral Density")

    n_coord = x_true.shape[-1]
    fig_by_coord_1d = plt.figure(figsize=(n_coord * 4, 6))
    for coord_ind, cname in enumerate(coord_names):
        ax = fig_by_coord_1d.add_subplot(n_coord, 1, coord_ind + 1)
        _plot_training_1d(ax, coord_ind, t_train, x_train, x_true, x_smooth, cname)

    fig_by_coord_1d.axes[-1].legend()

    return fig_composite, fig_by_coord_1d


def _plot_data_psd(
    ax: Axes,
    x: np.ndarray,
    coord_names: Sequence[str],
    traj_type: str = "train",
):
    """Plot the power spectral density of training data."""
    if traj_type == "train":
        color = COLOR.MEAS
    elif traj_type == "sim":
        color = COLOR.SIM
    elif traj_type == "true":
        color = COLOR.TRUE
    elif traj_type == "smooth":
        color = COLOR.EST
    else:
        raise ValueError(f"Unknown traj_type '{traj_type}'")
    coord_names = [name + f" {traj_type}" for name in coord_names]
    for coord, series in zip(coord_names, x.T):
        ax.loglog(
            np.abs(scipy.fft.rfft(series)) / np.sqrt(len(series)),
            color=color,
            label=coord,
        )
    ax.legend()
    ax.set(xlabel="Wavenumber")
    ax.set(ylabel="Magnitude")


def _plot_training_1d(
    ax: Axes,
    coord_ind: int,
    t_train: np.ndarray,
    x_train: np.ndarray,
    x_true: np.ndarray,
    x_smooth: Optional[np.ndarray],
    coord_name: str,
):
    ax.plot(t_train, x_train[..., coord_ind], ".", color=COLOR.MEAS, label="measured")
    ax.plot(t_train, x_true[..., coord_ind], "-", color=COLOR.TRUE, label="true")
    if x_smooth is not None:
        ax.plot(t_train, x_smooth[..., coord_ind], color=COLOR.EST, label="smoothed")
    ax.set(xlabel="t", ylabel=coord_name)


def _plot_pde_training_data(last_train, last_train_true, smoothed_last_train):
    """Plot training data (and smoothed training data, if different)."""
    # 1D:
    if len(last_train.shape) == 3:
        fig, axs = plt.subplots(1, 3, figsize=(18, 6))
        axs[0].imshow(last_train_true, vmin=0, vmax=last_train_true.max())
        axs[0].set(title="True Data")
        axs[1].imshow(last_train, vmin=0, vmax=last_train_true.max())
        axs[1].set(title="Noisy Data")
        axs[2].imshow(smoothed_last_train, vmin=0, vmax=last_train_true.max())
        axs[2].set(title="Smoothed Data")
        return plt.show()


def _plot_test_sim_data_1d_panel(
    axs: Sequence[Axes],
    x_true: Optional[np.ndarray],
    x_sim: np.ndarray,
    t_test: np.ndarray,
    t_sim: np.ndarray,
    coord_names: Sequence[str],
) -> None:
    for ordinate, ax in enumerate(axs):
        if x_true is not None:
            ax.plot(t_test, x_true[:, ordinate], color=COLOR.TRUE, label="True")
        axs[ordinate].plot(
            t_sim, x_sim[:, ordinate], "--", color=COLOR.SIM, label="Simulation"
        )
        axs[ordinate].set(xlabel="t", ylabel=coord_names[ordinate])


def _plot_test_sim_data_2d(
    ax: Axes,
    x_true: Optional[np.ndarray],
    x_sim: np.ndarray,
    labels: bool,
    coord_names: Sequence[str],
) -> None:
    if x_true is not None:
        ax.plot(x_true[:, 0], x_true[:, 1], color=COLOR.TRUE, label="True Values")
    ax.plot(x_sim[:, 0], x_sim[:, 1], "--", color=COLOR.SIM, label="Simulation")
    if labels:
        ax.set(xlabel=coord_names[0], ylabel=coord_names[1])
    else:
        ax.set(xticks=[], yticks=[])


def _plot_test_sim_data_3d(
    ax: Axes, x_vals: np.ndarray, label: Optional[str], coord_names: Sequence[str]
):
    if label == "True":
        color = COLOR.TRUE
    elif label == "Simulation":
        color = COLOR.SIM
    else:
        color = None
    ax.plot(x_vals[:, 0], x_vals[:, 1], x_vals[:, 2], color=color, label=label)
    if label:
        ax.set(xlabel=coord_names[0], ylabel=coord_names[1], zlabel=coord_names[2])
    else:
        ax.set(xticks=[], yticks=[], zticks=[])


def plot_test_trajectory(
    x_true: np.ndarray,
    x_sim: np.ndarray,
    t_test: np.ndarray,
    t_sim: np.ndarray,
    figs: Optional[tuple[Figure, Figure]] = None,
    coord_names: Optional[Sequence[str]] = None,
) -> tuple[Figure, Figure]:
    """Plot a test trajectory

    Args:
        last_test: a single trajectory of the system
        model: a trained model to simulate and compare to test data
        dt: the time interval in test data

    Returns:
        The sequence of axes used for the single-dimension time-series plots.
        If ``axs`` is provided, the same sequence is returned.
    """
    if coord_names is None:
        coord_names = [f"$x_{i}$" for i in range(x_true.shape[1])]
    if not figs:
        fig_by_coord_1d, axs_by_coord = plt.subplots(
            x_true.shape[1], 1, sharex=True, figsize=(7, 9)
        )
        if x_true.shape[1] == 2:
            fig_composite, axs_composite = plt.subplots(1, 2, figsize=(10, 4.5))
        elif x_true.shape[1] == 3:
            fig_composite, axs_composite = plt.subplots(
                1, 2, figsize=(10, 4.5), subplot_kw={"projection": "3d"}
            )
        else:
            raise ValueError("Can only plot 2d or 3d data.")
    else:
        fig_composite, fig_by_coord_1d = figs
        axs_composite = fig_composite.axes
        axs_by_coord = fig_by_coord_1d.axes

    assert isinstance(axs_composite, list)
    assert isinstance(axs_by_coord, list)
    _plot_test_sim_data_1d_panel(axs_by_coord, None, x_sim, t_test, t_sim, coord_names)
    axs_by_coord[-1].legend()
    if x_true.shape[1] == 2:
        _plot_test_sim_data_2d(
            axs_composite[0], None, x_sim, labels=True, coord_names=coord_names
        )
    elif x_true.shape[1] == 3:
        _plot_test_sim_data_3d(axs_composite[0], x_sim, "Simulation", coord_names)
    axs_composite[0].legend()
    _plot_data_psd(axs_composite[1], x_sim, coord_names, traj_type="sim")
    if not figs:
        fig_by_coord_1d.suptitle("Test Trajectories by Dimension")
        fig_composite.suptitle("Full Test Trajectories")
        axs_composite[0].set(title="true trajectory")
        axs_composite[0].set(title="model simulation")
    return fig_composite, fig_by_coord_1d



"""Utilities to convert a `dysts` dynamical system object's rhs to SymPy.

This module inspects the source of an object's RHS method (by default
named ``rhs``), parses the function using ``ast``, and converts the
returned expression(s) into SymPy expressions.

The conversion is intentionally conservative and aims to handle common
patterns used in simple rhs implementations, e.g. returning a tuple/list
of arithmetic expressions, using indexing into a state vector (``x[0]``),
and calls to common ``numpy``/``math`` functions (``np.sin``, ``math.exp``, ...).

Limitations:
- It does not execute arbitrary code from the inspected function.
- Complex control flow, loops, or non-trivial Python constructs may not
  be fully supported.

Example
-------
from dysts.flows import Lorenz
from inspect_to_sympy import object_to_sympy_rhs

lor = Lorenz()
symbols, exprs, lambda_rhs = object_to_sympy_rhs(lor)
# `symbols` is a list of SymPy symbols for the state vector
# `exprs` is a list of SymPy expressions for the RHS
# `lambda_rhs` is a SymPy Lambda mapping state symbols -> rhs expressions
"""

from __future__ import annotations

import ast
import inspect
import textwrap
from typing import Any, Callable, Dict, List, Tuple

import numpy as np
import sympy as sp
from dysts.base import BaseDyn


def _is_name(node: ast.AST, name: str) -> bool:
    return isinstance(node, ast.Name) and node.id == name


class _ASTToSympy(ast.NodeVisitor):
    def __init__(
        self,
        state_name: str,
        state_symbols: List[sp.Symbol],
        locals_map: Dict[str, Any],
    ):
        self.state_name = state_name
        self.state_symbols = state_symbols
        self.locals = dict(locals_map)

    def generic_visit(self, node):
        raise NotImplementedError(f"AST node not supported: {node!r}")

    def visit_Constant(self, node: ast.Constant):
        return sp.sympify(node.value)

    def visit_Num(self, node: ast.Num):
        return sp.sympify(node.n)

    def visit_Name(self, node: ast.Name):
        if node.id in self.locals:
            return self.locals[node.id]
        return sp.Symbol(node.id)

    def visit_Tuple(self, node: ast.Tuple):
        elems = []
        for elt in node.elts:
            val = self.visit(elt)
            if isinstance(val, (list, tuple)):
                elems.extend(list(val))
            else:
                elems.append(val)
        return tuple(elems)

    def visit_List(self, node: ast.List):
        elems = []
        for elt in node.elts:
            val = self.visit(elt)
            if isinstance(val, (list, tuple)):
                elems.extend(list(val))
            else:
                elems.append(val)
        return elems

    def visit_Starred(self, node: ast.Starred):
        # Handle starred expressions like `*x` in list/tuple literals.
        # If the starred value is the state vector name, expand to state symbols.
        if isinstance(node.value, ast.Name) and node.value.id == self.state_name:
            return tuple(self.state_symbols)
        # Otherwise, evaluate the value and if it is a sequence, return its items
        val = self.visit(node.value)
        if isinstance(val, (list, tuple)):
            return tuple(val)
        raise NotImplementedError(
            "Unsupported starred expression; cannot expand non-iterable"
        )

    def visit_BinOp(self, node: ast.BinOp):
        left = self.visit(node.left)
        right = self.visit(node.right)
        if isinstance(node.op, ast.Add):
            return left + right
        if isinstance(node.op, ast.Sub):
            return left - right
        if isinstance(node.op, ast.Mult):
            return left * right
        if isinstance(node.op, ast.Div):
            return left / right
        if isinstance(node.op, ast.Pow):
            return left**right
        if isinstance(node.op, ast.Mod):
            return left % right
        raise NotImplementedError(f"Binary op not supported: {node.op!r}")

    def visit_UnaryOp(self, node: ast.UnaryOp):
        operand = self.visit(node.operand)
        if isinstance(node.op, ast.USub):
            return -operand
        if isinstance(node.op, ast.UAdd):
            return +operand
        raise NotImplementedError(f"Unary op not supported: {node.op!r}")

    def visit_Call(self, node: ast.Call):
        # Determine function name
        func = node.func
        func_name = None
        mod_name = None

        if isinstance(func, ast.Name):
            func_name = func.id
        elif isinstance(func, ast.Attribute):
            # e.g. np.sin or math.exp
            if isinstance(func.value, ast.Name):
                mod_name = func.value.id
            func_name = func.attr
        else:
            raise NotImplementedError(
                f"Call to unsupported func node: {ast.dump(func)}"
            )

        # Map common numpy/math functions to sympy
        func_map = {
            "sin": sp.sin,
            "cos": sp.cos,
            "tan": sp.tan,
            "exp": sp.exp,
            "log": sp.log,
            "sqrt": sp.sqrt,
            "abs": sp.Abs,
            "atan": sp.atan,
            "asin": sp.asin,
            "acos": sp.acos,
        }

        args = [self.visit(a) for a in node.args]

        # Special-case array constructors: return underlying list/tuple
        if func_name in ("array", "asarray") and mod_name in ("np", "numpy"):
            # expect a single positional arg that's a list/tuple
            if len(args) == 1:
                return args[0]

        if func_name in func_map:
            return func_map[func_name](*args)

        # Unknown function: create a Sympy Function
        symf = sp.Function(func_name)
        return symf(*args)

    def visit_Subscript(self, node: ast.Subscript):
        # Support patterns like x[0] where x is the state vector name
        value = node.value
        # handle simple constant index
        if _is_name(value, self.state_name):
            # Python >=3.9: slice is directly the node.slice
            idx_node = node.slice
            if isinstance(idx_node, ast.Constant):
                idx = idx_node.value
            else:
                raise NotImplementedError(
                    "Only constant indices into state vector supported"
                )
            return self.state_symbols[idx]

        # If it's something else, try to evaluate generically
        base = self.visit(value)
        # slice may be constant
        if isinstance(node.slice, ast.Constant):
            key = node.slice.value
            return base[key]
        raise NotImplementedError("Unsupported subscript pattern")


def _numeric_consistency_check(
    dysts_flow: BaseDyn,
    rhsfunc: Callable,
    arg_names: List[str],
    state_names: List[str],
    vector_mode: bool,
    sys_dim: int,
    lambda_rhs: sp.Lambda,
) -> None:
    """Compare the original dysts rhs function to the SymPy-derived lambda.

    Raises a RuntimeError if they disagree.
    """
    # default to nonnegative support (e.g. Lotka volterra)
    random_state = np.random.standard_exponential(size=sys_dim)

    # Construct call arguments for the original function (bound method).
    call_args = []
    for name in arg_names:
        if name == "self":
            continue
        if name in state_names and not vector_mode:
            idx = state_names.index(name)
            call_args.append(random_state[idx])
        elif name in state_names and vector_mode:
            call_args.append(np.asarray(random_state, dtype=float))
        elif name == "t":
            call_args.append(float(np.random.standard_normal(size=())))
        else:
            call_args.append(dysts_flow.params[name])

    dysts_val = rhsfunc(*call_args)
    orig_arr = np.asarray(dysts_val, dtype=float).ravel()

    sym_val = lambda_rhs(*tuple(random_state))
    sym_arr = np.asarray(sym_val, dtype=float).ravel()

    if orig_arr.shape != sym_arr.shape:
        raise RuntimeError(
            f"_rhs shape {orig_arr.shape} != sympy shape {sym_arr.shape}"
        )

    if not np.allclose(orig_arr, sym_arr, rtol=1e-6, atol=1e-9):
        raise RuntimeError("Numeric mismatch between original and sympy conversion.")


def dynsys_to_sympy(
    obj: Any, func_name: str = "_rhs"
) -> Tuple[List[sp.Symbol], List[sp.Expr], sp.Lambda]:
    """Inspect ``obj`` for a method named ``func_name`` and return a SymPy
    representation of its RHS.

    Returns:
        a tuple ``(state_symbols, exprs, lambda_rhs)`` where ``state_symbols``
    is a list of SymPy symbols for the state vector, ``exprs`` is a list of
    SymPy expressions for the RHS components, and ``lambda_rhs`` is a SymPy
    Lambda mapping the state symbols to the RHS vector.

    Example:

    >>> from dysts.flows import Lorenz
    >>> from inspect_to_sympy import dynsys_to_sympy
    >>> lor = Lorenz()
    >>> symbols, exprs, lambda_rhs = dynsys_to_sympy(lor)
    >>> print(lor._rhs(1, 2, 3, t=0.0, **lor.params))
    (10, 23, -6.0009999999999994)

    >>> print(tuple(lambda_rhs(1, 2, 3)))
    (10, 23, -6.00100000000000)

    """

    if not hasattr(obj, func_name):
        raise AttributeError(f"Object has no attribute {func_name!r}")

    func = getattr(obj, func_name)
    src = inspect.getsource(func)
    src = textwrap.dedent(src)

    parsed = ast.parse(src)

    # Find first FunctionDef
    fndef = None
    for node in parsed.body:
        if isinstance(node, ast.FunctionDef):
            fndef = node
            break
    if fndef is None:
        raise RuntimeError("No function definition found in source")

    # Determine state argument names. Common dysts signature:
    # (self, *states, t, *parameters). Prefer obj.dimension when available.
    arg_names = [a.arg for a in fndef.args.args]
    if len(arg_names) == 0:
        raise RuntimeError("Function has no arguments")

    start_idx = 0
    if arg_names[0] == "self":
        start_idx = 1

    vector_mode = False
    state_args: List[str]
    t_idx = None
    if "t" in arg_names:
        t_idx = arg_names.index("t")

    if hasattr(obj, "dimension") and isinstance(getattr(obj, "dimension"), int):
        n_state = int(getattr(obj, "dimension"))
        if t_idx is not None:
            potential = arg_names[start_idx:t_idx]
            if len(potential) >= n_state:
                state_args = potential[:n_state]
            else:
                state_args = [arg_names[start_idx]]
                vector_mode = True
        else:
            potential = arg_names[start_idx:]
            if len(potential) >= n_state:
                state_args = potential[:n_state]
            else:
                state_args = [arg_names[start_idx]]
                vector_mode = True
    else:
        if t_idx is not None:
            state_args = arg_names[start_idx:t_idx]
            if len(state_args) == 0:
                state_args = [arg_names[start_idx]]
                vector_mode = True
            elif len(state_args) == 1:
                # single name could be vector or scalar; assume vector-mode
                vector_mode = True
        else:
            state_args = [arg_names[start_idx]]
            vector_mode = True

    # If vector_mode, inspect AST for subscript/index usage or tuple unpacking
    if vector_mode:
        state_name = state_args[0]
        max_index = -1
        unpack_size = None
        for node in ast.walk(fndef):
            if (
                isinstance(node, ast.Subscript)
                and isinstance(node.value, ast.Name)
                and node.value.id == state_name
            ):
                sl = node.slice
                if isinstance(sl, ast.Constant) and isinstance(sl.value, int):
                    if sl.value > max_index:
                        max_index = sl.value
            if isinstance(node, ast.Assign):
                if isinstance(node.value, ast.Name) and node.value.id == state_name:
                    targets = node.targets
                    if len(targets) == 1 and isinstance(
                        targets[0], (ast.Tuple, ast.List)
                    ):
                        unpack_size = len(targets[0].elts)

        if unpack_size is not None:
            n_state = unpack_size
        elif max_index >= 0:
            n_state = max_index + 1
        else:
            n_state = int(getattr(obj, "dimension", 3))

        state_symbols = [sp.Symbol(f"x{i}") for i in range(n_state)]
        primary_state_name = state_name
    else:
        # individual state args -> use their arg names as symbol names
        state_symbols = [sp.Symbol(n) for n in state_args]
        primary_state_name = state_args[0] if len(state_args) > 0 else "x"

    # Build locals mapping from known state arg names and parameters
    locals_map: Dict[str, Any] = {}
    for i, name in enumerate(state_args):
        if i < len(state_symbols):
            locals_map[name] = state_symbols[i]

    # map parameters (if present) to numeric values or symbols
    if hasattr(obj, "parameters") and isinstance(getattr(obj, "parameters"), dict):
        params = getattr(obj, "parameters")
        if t_idx is not None:
            param_arg_names = arg_names[t_idx + 1 :]
        else:
            param_arg_names = []
        for pname in param_arg_names:
            if pname in params:
                locals_map[pname] = sp.sympify(params[pname])
            else:
                locals_map[pname] = sp.Symbol(pname)

    converter = _ASTToSympy(primary_state_name, state_symbols, locals_map)

    return_expr = None
    # Walk through function body statements, handle Assign and Return
    for stmt in fndef.body:
        if isinstance(stmt, ast.Assign):
            # only simple single-target assignments supported
            if len(stmt.targets) != 1:
                raise ValueError("Only single-target assignments supported")
            target = stmt.targets[0]
            if isinstance(target, ast.Name):
                value_expr = converter.visit(stmt.value)
                locals_map[target.id] = value_expr
            elif (
                isinstance(target, (ast.Tuple, ast.List))
                and isinstance(stmt.value, ast.Name)
                and stmt.value.id == state_name
            ):
                # unpacking like a,b,c = x -> map names to state symbols
                for i, elt in enumerate(target.elts):
                    if isinstance(elt, ast.Name):
                        locals_map[elt.id] = state_symbols[i]
        elif isinstance(stmt, ast.Return):
            return_expr = stmt.value

    if return_expr is None:
        # maybe last statement is an Expr with list construction;
        # try to find a Return node deep
        for node in ast.walk(fndef):
            if isinstance(node, ast.Return):
                return_expr = node.value
                break

    if return_expr is None:
        raise RuntimeError("No return expression found in function body")

    # Refresh converter with updated locals
    converter = _ASTToSympy(primary_state_name, state_symbols, locals_map)
    rhs_val = converter.visit(return_expr)

    if isinstance(rhs_val, (list, tuple)):
        exprs = list(rhs_val)
    else:
        # single expression: treat as 1-dim RHS
        exprs = [rhs_val]

    lambda_rhs = sp.Lambda(tuple(state_symbols), sp.Matrix(exprs))

    # Run numeric consistency guard (raises on mismatch)
    _numeric_consistency_check(
        obj,
        func,
        arg_names,
        state_args,
        vector_mode,
        len(state_symbols),
        lambda_rhs,
    )

    return state_symbols, exprs, lambda_rhs


__all__ = ["dynsys_to_sympy"]



from logging import getLogger
from typing import Any, Callable, Optional, cast

import dysts.flows
import dysts.systems
import numpy as np
import scipy

from ._dysts_to_sympy import dynsys_to_sympy
from ._odes import SHO, CubicHO, Hopf, Kinematics, LotkaVolterra, VanDerPol
from ._plotting import plot_training_data
from ._typing import Float1D, ProbData
from ._utils import _sympy_expr_to_feat_coeff

try:
    import jax

    from ._diffrax_solver import _gen_data_jax
except ImportError:
    raise

INTEGRATOR_KEYWORDS = {"rtol": 1e-12, "method": "LSODA", "atol": 1e-12}
MOD_LOG = getLogger(__name__)

ODE_CLASSES = {
    klass.lower(): getattr(dysts.flows, klass)
    for klass in dysts.systems.get_attractor_list()
}
ODE_CLASSES.update(
    {
        "lotkavolterra": LotkaVolterra,
        "sho": SHO,
        "cubicho": CubicHO,
        "hopf": Hopf,
        "vanderpol": VanDerPol,
        "kinematics": Kinematics,
    }
)


def gen_data(
    system: str,
    seed: int,
    n_trajectories: int = 1,
    ic_stdev: float = 3,
    noise_abs: Optional[float] = None,
    noise_rel: Optional[float] = None,
    dt: float = 0.01,
    t_end: float = 10,
    display: bool = False,
    array_namespace: str = "numpy",
) -> dict[str, Any]:
    """Generate random training and test data

    An Experiment step according to the mitosis experiment runner.
    Note that test data has no noise.

    Arguments:
        system: the system to integrate
        seed (int): the random seed for number generation
        n_trajectories (int): number of trajectories of training data
        ic_stdev (float): standard deviation for generating initial conditions
        noise_abs (float): measurement noise standard deviation.
            Defaults to .1 if noise_rel is None.
        noise_rel (float): measurement noise-to-signal power ratio.
            Either noise_abs or noise_rel must be None.  Defaults to
            None.
        dt: time step for sample
        t_end: end time of simulation
        display: Whether to display graphics of generated data.

    Returns:
        dictionary of data and descriptive information
    """
    try:
        dyst_sys = ODE_CLASSES[system.lower()]()
    except KeyError as e:
        raise ValueError(
            f"Unknown system {system}.  Check {__name__}.ODE_CLASSES"
        ) from e
    input_features, sp_expr, sp_lambda = dynsys_to_sympy(dyst_sys)
    coeff_true = _sympy_expr_to_feat_coeff(sp_expr)
    rhsfunc = lambda t, X: dyst_sys.rhs(X, t)  # noqa: E731
    try:
        x0_center = dyst_sys.ic
    except KeyError:
        x0_center = np.zeros((len(input_features)), dtype=np.float64)
    try:
        nonnegative = getattr(dyst_sys, "nonnegative", False)
    except KeyError:
        nonnegative = False
    if noise_abs is not None and noise_rel is not None:
        raise ValueError("Cannot specify both noise_abs and noise_rel")
    elif noise_abs is None and noise_rel is None:
        noise_abs = 0.1

    MOD_LOG.info(f"Generating {n_trajectories} trajectories of f{system}")
    prob_data_list: list[ProbData] = []
    if array_namespace == "numpy":
        feature_names = [feat.name for feat in input_features]
        for _ in range(n_trajectories):
            seed += 1
            prob = _gen_data(
                rhsfunc,
                feature_names,
                seed,
                x0_center=x0_center,
                nonnegative=nonnegative,
                ic_stdev=ic_stdev,
                noise_abs=noise_abs,
                noise_rel=noise_rel,
                dt=dt,
                t_end=t_end,
            )
            prob_data_list.append(prob)
    elif array_namespace == "jax":
        try:
            globals()["_gen_data_jax"]
        except KeyError:
            raise ImportError(
                "jax data generation requested but diffrax or sympy2jax not"
                " installed"
            )
        this_seed = jax.random.PRNGKey(seed)
        for _ in range(n_trajectories):
            this_seed, _ = jax.random.split(this_seed)
            prob = _gen_data_jax(
                sp_expr,
                input_features,
                this_seed,
                x0_center=x0_center,
                nonnegative=nonnegative,
                ic_stdev=ic_stdev,
                noise_abs=noise_abs,
                noise_rel=noise_rel,
                dt=dt,
                t_end=t_end,
            )
            prob_data_list.append(prob)
    else:
        raise ValueError(
            f"Unknown array_namespace {array_namespace}.  Must be 'numpy' or 'jax'"
        )
    if display and prob_data_list:
        sample = prob_data_list[0]
        figs = plot_training_data(sample.t_train, sample.x_train, sample.x_train_true)
        figs[0].suptitle("Sample Trajectory")
    return {
        "data": {"trajectories": prob_data_list, "coeff_true": coeff_true},
        "main": f"{n_trajectories} trajectories of {rhsfunc}",
        "metrics": {"rel_noise": noise_rel, "abs_noise": noise_abs},
    }


def _gen_data(
    rhs_func: Callable,
    input_features: list[str],
    seed: Optional[int],
    x0_center: Float1D,
    ic_stdev: float,
    noise_abs: Optional[float],
    noise_rel: Optional[float],
    nonnegative: bool,
    dt: float,
    t_end: float,
) -> ProbData:
    rng = np.random.default_rng(seed)
    t_train = np.arange(0, t_end, dt)
    t_train_span = (t_train[0], t_train[-1])
    if nonnegative:
        shape = ((x0_center + 1) / ic_stdev) ** 2
        scale = ic_stdev**2 / (x0_center + 1)
        x0 = np.array([rng.gamma(k, theta) for k, theta in zip(shape, scale)])
    else:
        x0 = ic_stdev * rng.standard_normal(len(input_features)) + x0_center
    x_train = scipy.integrate.solve_ivp(
        rhs_func,
        t_train_span,
        x0,
        t_eval=t_train,
        **INTEGRATOR_KEYWORDS,
    ).y.T

    x_train_true = np.copy(x_train)
    x_train_true_dot = np.array([rhs_func(0, xi) for xi in x_train_true])
    if noise_rel is not None:
        noise_abs = np.sqrt(_signal_avg_power(x_train) * noise_rel)
    x_train = x_train + cast(float, noise_abs) * rng.standard_normal(x_train.shape)

    return ProbData(
        dt, t_train, x_train, x_train_true, x_train_true_dot, input_features
    )


def _max_amplitude(signal: np.ndarray, axis: int) -> float:
    return np.abs(scipy.fft.rfft(signal, axis=axis)[1:]).max() / np.sqrt(
        signal.shape[axis]
    )


def _signal_avg_power(signal: np.ndarray) -> float:
    return np.square(signal).mean()



from typing import Optional

import diffrax
import jax
import jax.numpy as jnp
import sympy2jax
from sympy import Expr, Symbol

from ._typing import ProbData

jax.config.update("jax_enable_x64", True)


def _gen_data_jax(
    exprs: list[Expr],
    input_features: list[Symbol],
    seed: jax.Array,
    x0_center: jax.Array,
    ic_stdev: float,
    noise_abs: Optional[float],
    noise_rel: Optional[float],
    nonnegative: bool,
    dt: float,
    t_end: float,
) -> ProbData:
    rhstree = sympy2jax.SymbolicModule(exprs)

    def ode_sys(t, state, args):
        return jnp.asarray(
            rhstree(
                **{
                    str(x_sym): state_i
                    for x_sym, state_i in zip(input_features, state, strict=True)
                }
            )
        )

    term = diffrax.ODETerm(ode_sys)
    solver = diffrax.Tsit5()
    save_at = diffrax.SaveAt(ts=jnp.arange(0, t_end, dt), dense=True)

    # Random initialization
    key, subkey = jax.random.split(seed)
    t_train = jnp.arange(0, t_end, dt)
    if nonnegative:
        shape = ((x0_center + 1) / ic_stdev) ** 2
        scale = ic_stdev**2 / (x0_center + 1)
        x0 = jnp.array(
            jax.random.gamma(subkey, k) * theta for k, theta in zip(shape, scale)
        ).T

    else:
        x0 = ic_stdev * jax.random.normal(subkey, (len(input_features),)) + x0_center
    key, subkey = jax.random.split(key)

    # IVPs
    sol = diffrax.diffeqsolve(
        term,
        solver,
        t0=0,
        t1=t_end,
        dt0=dt,  # Initial step size
        y0=x0,
        args=(),
        saveat=save_at,
        max_steps=int(10 * (t_end - 0) / dt),
    )
    x_train_true: jax.Array = sol.ys  # type: ignore

    # Measurement noise
    if noise_abs is None:
        assert noise_rel is not None  # force type narrowing
        noise_abs = float(jnp.sqrt(_signal_avg_power(x_train_true)) * noise_rel)

    x_train = x_train_true + jax.random.normal(key, x_train_true.shape) * noise_abs

    # True Derivatives
    x_train_true_dot = jnp.array([ode_sys(0, xi, None) for xi in x_train_true])

    stringy_features = [sym.name for sym in input_features]
    return ProbData(
        dt, t_train, x_train, x_train_true, x_train_true_dot, stringy_features, sol
    )


def _signal_avg_power(signal: jax.Array) -> jax.Array:
    return jnp.square(signal).mean()


## %  # noqa:E266
if __name__ == "__main__":
    # Debug example
    from sindy_exp._data import gen_data

    data_dict = gen_data(
        "valliselnino",
        seed=50,
        n_trajectories=1,
        ic_stdev=3,
        noise_rel=0.1,
        display=True,
        array_namespace="jax",
    )
    print(data_dict["input_features"])



from importlib import resources
from logging import getLogger
from typing import Callable, TypeVar

import matplotlib.pyplot as plt
import numpy as np
import pysindy as ps
import sympy as sp
from dysts.base import DynSys

from ._plotting import (
    compare_coefficient_plots_from_dicts,
    plot_test_trajectory,
    plot_training_data,
)
from ._typing import (
    DynamicsTrialData,
    FullDynamicsTrialData,
    ProbData,
    SINDyTrialUpdate,
    _BaseSINDy,
)
from ._utils import (
    _simulate_test_data,
    coeff_metrics,
    integration_metrics,
    unionize_coeff_dicts,
)

metric_ordering = {
    "coeff_precision": "max",
    "coeff_f1": "max",
    "coeff_recall": "max",
    "coeff_mae": "min",
    "coeff_mse": "min",
    "mse_plot": "min",
    "mae_plot": "min",
}


T = TypeVar("T", bound=int)
DType = TypeVar("DType", bound=np.dtype)
MOD_LOG = getLogger(__name__)
LOCAL_DYNAMICS_PATH = resources.files("sindy_exp").joinpath("addl_attractors.json")


def _add_forcing(
    forcing_func: Callable[[float], np.ndarray[tuple[T], DType]],
    auto_func: Callable[
        [float, np.ndarray[tuple[T], DType]], np.ndarray[tuple[T], DType]
    ],
) -> Callable[[float, np.ndarray], np.ndarray]:
    """Add a time-dependent forcing term to a rhs func

    Args:
        forcing_func: The forcing function to add
        auto_func: An existing rhs func for solve_ivp

    Returns:
        A rhs function for integration
    """

    def sum_of_terms(
        t: float, state: np.ndarray[tuple[T], DType]
    ) -> np.ndarray[tuple[T], DType]:
        return np.array(forcing_func(t)) + np.array(auto_func(t, state))

    return sum_of_terms


class LotkaVolterra(DynSys):
    """Lotka-Volterra (predator-prey) dynamical system."""

    nonnegative = True

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)

    @staticmethod
    def _rhs(x, y, t: float, alpha, beta, gamma, delta) -> np.ndarray:
        """LV dynamics

        Args:
            x: prey population
            y: predator population
            t: time (ignored, since autonomous)
            alpha: prey growth rate
            beta: predation rate
            delta: predator reproduction rate
            gamma: predator death rate
        """
        dxdt = alpha * x - beta * x * y
        dydt = delta * x * y - gamma * y
        return np.array([dxdt, dydt])


class Hopf(DynSys):
    """Hopf normal form dynamical system."""

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)

    @staticmethod
    def _rhs(x, y, t: float, mu, omega, A) -> np.ndarray:
        dxdt = mu * x - omega * y - A * (x**3 + x * y**2)
        dydt = omega * x + mu * y - A * (x**2 * y + y**3)
        return np.array([dxdt, dydt])


class SHO(DynSys):
    """Linear damped simple harmonic oscillator"""

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)

    @staticmethod
    def _rhs(x, y, t: float, a, b, c, d) -> np.ndarray:
        dxdt = a * x + b * y
        dydt = c * x + d * y
        return np.array([dxdt, dydt])


class CubicHO(DynSys):
    """Cubic damped harmonic oscillator."""

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)

    @staticmethod
    def _rhs(x, y, t: float, a, b, c, d) -> np.ndarray:
        dxdt = a * x**3 + b * y**3
        dydt = c * x**3 + d * y**3
        return np.array([dxdt, dydt])


class VanDerPol(DynSys):
    """Van der Pol oscillator.

    dx/dt = y
    dy/dt = mu * (1 - x^2) * y - x
    """

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)

    @staticmethod
    def _rhs(x, x_dot, t: float, mu) -> np.ndarray:
        dxdt = x_dot
        dx2dt2 = mu * (1 - x**2) * x_dot - x
        return np.array([dxdt, dx2dt2])


class Kinematics(DynSys):
    """One-dimensional kinematics with constant acceleration.

    dx/dt = v
    dv/dt = a
    """

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)

    @staticmethod
    def _rhs(x, v, t: float, a) -> np.ndarray:
        dxdt = v
        dvdt = a
        return np.array([dxdt, dvdt])


def fit_eval(
    data: tuple[list[ProbData], list[dict[sp.Expr, float]]],
    model: _BaseSINDy,
    simulations: bool = True,
    display: bool = True,
    return_all: bool = True,
) -> dict | tuple[dict, DynamicsTrialData | FullDynamicsTrialData]:
    """Fit and evaluate a SINDy model on a set of trajectories.

    Args:
        data: Tuple of (trajectories, true_equations), where ``trajectories`` is
            a list of ProbData objects and ``true_equations`` is a list of
            dictionaries mapping SymPy symbols to their true coefficients for
            each state coordinate.
        model: A SINDy-like model implementing the _BaseSINDy protocol.
        simulations: Whether to run forward simulations for evaluation.
        display: Whether to generate plots as part of evaluation.
        return_all: If True, return a dictionary containing metrics and the
            assembled DynamicsTrialData; otherwise return only the metrics
            dictionary.
    """

    trajectories, true_equations = data
    input_features = trajectories[0].input_features

    x_train = [traj.x_train for traj in trajectories]
    t_train = [traj.t_train for traj in trajectories]
    model.fit(x_train, t=t_train, feature_names=input_features)

    MOD_LOG.info(f"Fitting a model: {model}")
    coeff_true_dicts, coeff_est_dicts = unionize_coeff_dicts(model, true_equations)
    if isinstance(model.feature_library, ps.WeakPDELibrary):
        # WeakPDE library fails to simulate, so insert nonweak library
        # to Pipeline and SINDy model.
        inner_lib = model.feature_library.function_library
        model.feature_library = inner_lib  # type: ignore  # TODO: Fix in pysindy
    if isinstance(model, ps.SINDy) and hasattr(
        model.differentiation_method, "smoothed_x_"
    ):
        smooth_x = []
        for traj in trajectories:
            model.differentiation_method(traj.x_train, t=traj.t_train)
            smooth_x.append(model.differentiation_method.smoothed_x_)
    else:  # using WeakPDELibrary
        smooth_x = x_train
    trial_data = DynamicsTrialData(
        trajectories=trajectories,
        true_equations=coeff_true_dicts,
        sindy_equations=coeff_est_dicts,
        model=model,
        input_features=input_features,
        smooth_train=smooth_x,
    )
    MOD_LOG.info(f"Evaluating a model: {model}")
    metrics = coeff_metrics(coeff_est_dicts, coeff_true_dicts)
    if simulations:
        sims: list[SINDyTrialUpdate] = []
        integration_metric_list: list[dict[str, float | np.floating]] = []
        for traj in trajectories:
            sim = _simulate_test_data(model, traj.dt, traj.x_train_true)
            sims.append(sim)
            integration_metric_list.append(
                integration_metrics(
                    model,
                    traj.x_train_true,
                    traj.t_train,
                    traj.x_train_true_dot,
                )
            )

        agg_integration_metrics: dict[str, float | np.floating] = {}
        for key in integration_metric_list[0].keys():
            values = [m[key] for m in integration_metric_list]
            agg_integration_metrics[key] = float(np.mean(values))
        metrics.update(agg_integration_metrics)

        trial_data = FullDynamicsTrialData(sims=sims, **trial_data.__dict__)
    if display:
        plot_ode_panel(trial_data)
        for i, traj in enumerate(trajectories):
            fig_composite, fig_by_coord_1d = plot_training_data(
                traj.t_train,
                traj.x_train,
                traj.x_train_true,
                x_smooth=smooth_x[i],
                coord_names=input_features,
            )
            if simulations:
                # Overlay test trajectory time series on the coordinate-wise figure
                plot_test_trajectory(
                    traj.x_train_true,
                    sims[i].x_sim,
                    traj.t_train,
                    sims[i].t_sim,
                    figs=(fig_composite, fig_by_coord_1d),
                    coord_names=input_features,
                )
    if return_all:
        return {"metrics": metrics, "data": trial_data, "main": metrics["main"]}
    return metrics


def plot_ode_panel(trial_data: DynamicsTrialData):
    trial_data.model.print()
    compare_coefficient_plots_from_dicts(
        trial_data.sindy_equations,
        trial_data.true_equations,
        input_features=[_texify(feat) for feat in trial_data.input_features],
    )
    plt.show()


def _texify(input: str) -> str:
    if input[0] != "$":
        input = "$" + input
    if input[-1] != "$":
        input = input + "$"
    return input



import logging
from itertools import chain
from typing import cast
from warnings import warn

import numpy as np
import pysindy as ps
import sklearn
import sklearn.metrics
import sympy as sp
from sympy.parsing.sympy_parser import (
    convert_xor,
    implicit_multiplication_application,
    parse_expr,
    standard_transformations,
)

from sindy_exp._typing import SINDyTrialUpdate

from ._typing import Float1D, Float2D, _BaseSINDy

logger = logging.getLogger(__name__)


def _sympy_expr_to_feat_coeff(sp_expr: list[sp.Expr]) -> list[dict[sp.Expr, float]]:
    """Convert symbolic rhs expressions into feature/coeff dictionaries.

    Each expression is assumed to be a sum of terms; each term is either a
    simple SymPy expression or a product of a numeric coefficient and a
    feature expression. The output is a list of dictionaries, one per
    expression, mapping feature expressions to their numeric coefficients.
    """

    expressions: list[dict[sp.Expr, float]] = []

    def kv_term(term: sp.Expr) -> tuple[sp.Expr, float]:
        if not isinstance(term, sp.Mul):
            # Term is either a constant or a feature without a coefficient
            if isinstance(term, sp.Number):
                coeff = float(term)
                feat = sp.Integer(1)
            elif isinstance(term, (sp.Symbol, sp.Pow, sp.Function)):
                coeff = 1.0
                feat = term
            else:
                raise ValueError(f"Unrecognized term format: {term}")
        else:
            try:
                # Assume multiplication is coefficient * feature(s)
                coeff = float(term.args[0])
                args = term.args[1:]
            except TypeError:
                # e.g. x**2 or x * y has no numeric coefficient
                coeff = 1.0
                args = term.args
            if len(args) == 1:
                feat = args[0]
            else:
                feat = sp.Mul(*args)
        return feat, coeff

    for exp in sp_expr:
        expr_dict: dict[sp.Expr, float] = {}
        if not isinstance(exp, sp.Add):
            feat, coeff = kv_term(exp)
            expr_dict[feat] = coeff
        else:
            for term in exp.args:
                feat, coeff = kv_term(term)
                expr_dict[feat] = coeff

        expressions.append(expr_dict)
    return expressions


def _sindy_equations_to_sympy(model: _BaseSINDy) -> tuple[list[sp.Expr], list[sp.Expr]]:
    """Convert a SINDy model's string equations to SymPy expressions.

    Uses sympy's parser with ``convert_xor`` so that terms like ``x^2`` are
    interpreted as powers rather than bitwise XOR.
    """

    # Use a fixed precision for reproducible string equations.
    input_features = {feat: sp.sympify(feat) for feat in model.feature_names}
    eq_strings = model.equations(10)  # type: ignore[call-arg]
    feat_strs = model.feature_library.get_feature_names(
        input_features=model.feature_names
    )
    xforms = standard_transformations + (
        implicit_multiplication_application,
        convert_xor,
    )
    feat_symb = [
        parse_expr(fstr, transformations=xforms, evaluate=False) for fstr in feat_strs
    ]
    eq_symb = [
        parse_expr(
            eq, local_dict=input_features, transformations=xforms, evaluate=False
        )
        for eq in eq_strings
    ]
    return feat_symb, eq_symb


def diff_lookup(kind):
    normalized_kind = kind.lower().replace(" ", "")
    if normalized_kind == "finitedifference":
        return ps.FiniteDifference
    if normalized_kind == "smoothedfinitedifference":
        return ps.SmoothedFiniteDifference
    elif normalized_kind == "sindy":
        return ps.SINDyDerivative
    else:
        raise ValueError


def feature_lookup(kind):
    normalized_kind = kind.lower().replace(" ", "")
    if normalized_kind is None:
        return ps.PolynomialLibrary
    elif normalized_kind == "polynomial":
        return ps.PolynomialLibrary
    elif normalized_kind == "fourier":
        return ps.FourierLibrary
    elif normalized_kind == "weak":
        return ps.WeakPDELibrary
    elif normalized_kind == "pde":
        return ps.PDELibrary
    else:
        raise ValueError


def opt_lookup(kind):
    normalized_kind = kind.lower().replace(" ", "")
    if normalized_kind == "stlsq":
        return ps.STLSQ
    elif normalized_kind == "sr3":
        return ps.SR3
    elif normalized_kind == "miosr":
        return ps.MIOSR
    elif normalized_kind == "trap":
        return ps.TrappingSR3
    elif normalized_kind == "ensemble":
        return ps.EnsembleOptimizer
    else:
        raise ValueError


def coeff_metrics(
    coeff_est_dicts: list[dict[sp.Expr, float]],
    coeff_true_dicts: list[dict[sp.Expr, float]],
) -> dict[str, float | np.floating]:
    """Compute coefficient metrics from aligned coefficient dictionaries.

    Both arguments are expected to be lists of coefficient dictionaries sharing
    the same SymPy-expression keys, such as the output of ``unionize_coeff_dicts``.
    """

    if not coeff_true_dicts or not coeff_est_dicts:
        raise ValueError("Coefficient dictionaries must be non-empty")
    if len(coeff_true_dicts) != len(coeff_est_dicts):
        raise ValueError("True and estimated coefficients must have same length")

    features = list(coeff_true_dicts[0].keys())
    n_coord = len(coeff_true_dicts)
    n_feat = len(features)

    coeff_true = np.zeros((n_coord, n_feat), dtype=float)
    coefficients = np.zeros_like(coeff_true)

    for row_ind, (true_row, est_row) in enumerate(
        zip(coeff_true_dicts, coeff_est_dicts)
    ):
        if set(true_row.keys()) != set(features) or set(est_row.keys()) != set(
            features
        ):
            raise ValueError(
                "Coefficient dictionaries are not aligned across coordinates"
            )
        for col_ind, feat in enumerate(features):
            coeff_true[row_ind, col_ind] = true_row[feat]
            coefficients[row_ind, col_ind] = est_row[feat]

    metrics: dict[str, float | np.floating] = {}
    metrics["coeff_precision"] = sklearn.metrics.precision_score(
        coeff_true.flatten() != 0, coefficients.flatten() != 0
    )
    metrics["coeff_recall"] = sklearn.metrics.recall_score(
        coeff_true.flatten() != 0, coefficients.flatten() != 0
    )
    metrics["coeff_f1"] = sklearn.metrics.f1_score(
        coeff_true.flatten() != 0, coefficients.flatten() != 0
    )
    metrics["coeff_mse"] = sklearn.metrics.mean_squared_error(
        coeff_true.flatten(), coefficients.flatten()
    )
    metrics["coeff_mae"] = sklearn.metrics.mean_absolute_error(
        coeff_true.flatten(), coefficients.flatten()
    )
    metrics["main"] = metrics["coeff_f1"]
    return metrics


def pred_metrics(
    model: _BaseSINDy, x_test: np.ndarray, x_dot_test: np.ndarray
) -> dict[str, np.ndarray | float | np.floating]:
    preds = model.predict(x_test)
    err = preds - x_dot_test
    return {
        "pred_l2_fro": (np.linalg.norm(err) / np.linalg.norm(x_dot_test)),
        "pred_l2_each": (np.linalg.norm(err) / np.linalg.norm(x_dot_test)),
        "pred_r2": sklearn.metrics.r2_score(x_dot_test, preds),
    }


def integration_metrics(model: _BaseSINDy, x_test, t_train, x_dot_test):
    metrics = {}
    metrics["mse-plot"] = model.score(
        x_test,
        t_train,
        x_dot_test,
        metric=sklearn.metrics.mean_squared_error,
    )
    metrics["mae-plot"] = model.score(
        x_test,
        t_train,
        x_dot_test,
        metric=sklearn.metrics.mean_absolute_error,
    )
    return metrics


def unionize_coeff_dicts(
    model: _BaseSINDy,
    true_equations: list[dict[sp.Expr, float]],
) -> tuple[list[dict[sp.Expr, float]], list[dict[sp.Expr, float]]]:
    """Align true and estimated coefficient dictionaries using SymPy expressions.

    This function compares the symbolic features present in the true
    equations and in the fitted SINDy model's equations and returns two
    lists of coefficient dictionaries with identical feature keys. Missing
    features in either source are filled with a coefficient of 0.0.

    Args:
        model: Fitted SINDy-like model.
        true_equations: List of coefficient dictionaries mapping SymPy
            expressions to true coefficients, one dict per state coordinate.

    Returns:
        A pair ``(true_aligned, est_aligned)`` where each element is a list
        of coefficient dictionaries ``dict[sp.Expr, float]`` with identical
        keys across all coordinates.
    """

    feat_exprs, est_eq_exprs = _sindy_equations_to_sympy(model)
    # est_equations should be unnecessary; all terms are provided by features
    est_equations = _sympy_expr_to_feat_coeff(est_eq_exprs)

    if len(est_equations) != len(true_equations):
        raise ValueError(
            "True equations and estimated equations must have"
            " the same number of coordinates"
        )

    true_aligned: list[dict[sp.Expr, float]] = []
    est_aligned: list[dict[sp.Expr, float]] = []
    empty_feats: list[dict[sp.Expr, float]] = [dict.fromkeys(feat_exprs, 0.0)]

    all_features = [
        expr
        for eq in chain(empty_feats, true_equations, est_equations)
        for expr in eq.keys()
    ]
    all_features = list(dict.fromkeys(all_features))  # deduplicate, preserve order
    for true_eq, est_eq in zip(true_equations, est_equations):
        true_aligned.append({feat: true_eq.get(feat, 0.0) for feat in all_features})
        est_aligned.append({feat: est_eq.get(feat, 0.0) for feat in all_features})

    return true_aligned, est_aligned


def make_model(
    input_features: list[str],
    dt: float,
    diff_params: dict | ps.BaseDifferentiation,
    feat_params: dict | ps.feature_library.base.BaseFeatureLibrary,
    opt_params: dict | ps.BaseOptimizer,
) -> ps.SINDy:
    """Build a model with object parameters dictionaries

    e.g. {"kind": "finitedifference"} instead of FiniteDifference()
    """

    def finalize_param(lookup_func, pdict, lookup_key):
        try:
            cls_name = pdict.pop(lookup_key)
        except AttributeError:
            cls_name = pdict.vals.pop(lookup_key)
            pdict = pdict.vals

        param_cls = lookup_func(cls_name)
        param_final = param_cls(**pdict)
        pdict[lookup_key] = cls_name
        return param_final

    if isinstance(diff_params, ps.BaseDifferentiation):
        diff = diff_params
    else:
        diff = finalize_param(diff_lookup, diff_params, "diffcls")
    if isinstance(feat_params, ps.feature_library.base.BaseFeatureLibrary):
        features = feat_params
    else:
        features = finalize_param(feature_lookup, feat_params, "featcls")
    if isinstance(opt_params, ps.BaseOptimizer):
        opt = opt_params
    else:
        opt = finalize_param(opt_lookup, opt_params, "optcls")
    return ps.SINDy(
        differentiation_method=diff,
        optimizer=opt,
        t_default=dt,  # type: ignore
        feature_library=features,
        feature_names=input_features,
    )


def _simulate_test_data(
    model: _BaseSINDy, dt: float, x_test: Float2D
) -> SINDyTrialUpdate:
    """Add simulation data to grid_data

    This includes the t_sim and x_sim keys.  Does not mutate argument.
    Returns:
        Complete GridPointData
    """
    t_test = cast(Float1D, np.arange(0, len(x_test) * dt, step=dt))
    t_sim = t_test
    try:

        def quit(t, x):
            return np.abs(x).max() - 1000

        quit.terminal = True  # type: ignore
        x_sim = cast(
            Float2D,
            model.simulate(
                x_test[0],
                t_test,
                integrator_kws={
                    "method": "LSODA",
                    "rtol": 1e-12,
                    "atol": 1e-12,
                    "events": [quit],
                },
            ),
        )
    except ValueError:
        warn(message="Simulation blew up; returning zeros")
        x_sim = np.zeros_like(x_test)
    # truncate if integration returns wrong number of points
    t_sim = cast(Float1D, t_test[: len(x_sim)])
    return SINDyTrialUpdate(t_sim=t_sim, t_test=t_test, x_sim=x_sim)


def _drop_and_warn(arrs):
    """Drop trajectories that blew up during simulation"""
    maxlen = max(arr.shape[0] for arr in arrs)

    def _alert_short(arr):
        if arr.shape[0] < maxlen:
            warn(message="Dropping simulation due to blow-up")
            return False
        return True

    arrs = list(filter(_alert_short, arrs))
    if len(arrs) == 0:
        raise ValueError(
            "Simulations failed due to blow-up.  System is too stiff for solver's"
            " numerical tolerance"
        )
    return arrs



import pytest

try:
    from dysts.flows import Lorenz
except Exception:  # pragma: no cover - skip if dysts not installed
    pytest.skip(
        "dysts not available; skipping inspect_to_sympy tests", allow_module_level=True
    )

from sindy_exp._dysts_to_sympy import dynsys_to_sympy


def test_lorenz_to_sympy():
    lor = Lorenz()
    symbols, exprs, lambda_rhs = dynsys_to_sympy(lor, func_name="_rhs")
    assert len(symbols) == lor.dimension
    # evaluate lambda with simple numeric values
    vals = tuple(float(i + 1) for i in range(lor.dimension))
    mat = lambda_rhs(*vals)
    assert mat.shape[0] == lor.dimension



from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Literal,
    NamedTuple,
    Optional,
    Protocol,
    TypeVar,
    overload,
)

import numpy as np
import sympy as sp
from numpy.typing import NBitBase
from sympy import Expr
from typing_extensions import Self

NpFlt = np.dtype[np.floating]
Float1D = np.ndarray[tuple[int], NpFlt]
Float2D = np.ndarray[tuple[int, int], NpFlt]
Shape = TypeVar("Shape", bound=tuple[int, ...])
FloatND = np.ndarray[Shape, np.dtype[np.floating[NBitBase]]]


TrajectoryType = TypeVar("TrajectoryType", list[np.ndarray], np.ndarray)


class _BaseSINDy(Protocol):
    optimizer: Any
    feature_library: Any
    feature_names: list[str]

    def fit(self, x: TrajectoryType, t: TrajectoryType, *args, **kwargs) -> Self: ...

    def simulate(self, x0: np.ndarray, t: np.ndarray, **kwargs) -> np.ndarray: ...

    def score(
        self,
        x: TrajectoryType,
        t: TrajectoryType,
        x_dot: TrajectoryType,
        metric: Callable,
    ) -> float: ...

    def predict(self, x: np.ndarray, u: None | np.ndarray = None) -> np.ndarray: ...

    def coefficients(self): ...

    @overload
    def equations(self) -> list[str]: ...

    @overload
    def equations(self, precision: int) -> list[str]: ...

    @overload
    def equations(self, precision: int, fmt: Literal["str"] | None) -> list[str]: ...

    @overload
    def equations(
        self, precision: int, fmt: Literal["sympy"]
    ) -> list[dict[Expr, float]]: ...

    def print(self, precision: int, **kwargs) -> None: ...

    def get_feature_names(self) -> list[str]: ...


class ProbData(NamedTuple):
    """Data bundle for a single trajectory.

    Represents a trajectory's training data and associated metadata.
    """

    dt: float
    t_train: Float1D
    x_train: Float2D
    x_train_true: Float2D
    x_train_true_dot: Float2D
    input_features: list[str]
    integrator: Optional[Any] = None  # diffrax.Solution


class NestedDict(defaultdict):
    """A dictionary that splits all keys by ".", creating a sub-dict.

    Args: see superclass

    Example:

        >>> foo = NestedDict("a.b"=1)
        >>> foo["a.c"] = 2
        >>> foo["a"]["b"]
        1
    """

    def __missing__(self, key):
        try:
            prefix, subkey = key.split(".", 1)
        except ValueError:
            raise KeyError(key)
        return self[prefix][subkey]

    def __setitem__(self, key, value):
        if "." in key:
            prefix, suffix = key.split(".", 1)
            if self.get(prefix) is None:
                self[prefix] = NestedDict()
            return self[prefix].__setitem__(suffix, value)
        else:
            return super().__setitem__(key, value)

    def update(self, other: dict):  # type: ignore
        try:
            for k, v in other.items():
                self.__setitem__(k, v)
        except:  # noqa: E722
            super().update(other)

    def flatten(self):
        """Flattens a nested dictionary without mutating.  Returns new dict"""

        def _flatten(nested_d: dict) -> dict:
            new = {}
            for key, value in nested_d.items():
                if not isinstance(key, str):
                    raise TypeError("Only string keys allowed in flattening")
                if not isinstance(value, dict):
                    new[key] = value
                    continue
                for sub_key, sub_value in _flatten(value).items():
                    new[key + "." + sub_key] = sub_value
            return new

        return _flatten(self)


@dataclass
class DynamicsTrialData:
    trajectories: list[ProbData]
    true_equations: list[dict[sp.Expr, float]]
    sindy_equations: list[dict[sp.Expr, float]]
    model: _BaseSINDy
    input_features: list[str]
    smooth_train: list[np.ndarray]


@dataclass
class SINDyTrialUpdate:
    t_sim: Float1D
    t_test: Float1D
    x_sim: FloatND


@dataclass
class FullDynamicsTrialData(DynamicsTrialData):
    sims: list[SINDyTrialUpdate]
